import pandas as pd
import matplotlib.pyplot as plt

class PCSHOW:
    def __init__(self,data:pd.DataFrame,anno:pd.DataFrame=None):
        self.data = data
        self.anno = anno
        pass
    def pcplot(self,x:str, y:str, anno:pd.DataFrame=None,group:str=None,group_order:list=None,color:dict=None,size:dict=None,alpha:dict=None,marker:dict=None,ax:plt.Axes=None,**kwargs):
        anno = self.anno if anno is None else anno
        ax = ax if ax is not None else plt.gca()
        if self.anno is not None or group is not None:
            data = pd.concat([self.data,self.anno[group]],axis=1).dropna()
            groups = self.anno[group].unique() if group_order is None else group_order
            size = dict(zip(groups,[4 for i in groups])) if size is None else size # scatter size
            alpha = dict(zip(groups,[1 for i in groups])) if alpha is None else alpha # scatter size
            color = dict(zip(groups,['blue' for i in groups])) if color is None else color # scatter size
            marker = dict(zip(groups,['o' for i in groups])) if marker is None else marker # scatter size
            for g in groups:
                data_ = data[data[group]==g]
                ax.scatter(x=data_[x],y=data_[y],s=size[g],alpha=alpha[g],marker=marker[g],color=color[g],label=g,**kwargs)
        else:
            data = self.data
            ax.scatter(x=data[x],y=data[y],**kwargs)
        return
    def text_anno(self,x:str, y:str, anno:pd.DataFrame=None,anno_tag:str=None,ax:plt.Axes=None,**kwargs):
        from adjustText import adjust_text
        anno = self.anno if anno is None else anno
        ax = ax if ax is not None else plt.gca()
        data = pd.concat([self.data,anno[anno_tag]],axis=1).dropna()
        texts = []
        for ind,i in enumerate(data.index):
            texts.append(ax.text(data.loc[i][x],data.loc[i][y],data.loc[i][anno_tag],**kwargs))
        adjust_text(
            texts, 
            arrowprops=dict(arrowstyle='->', color='gray', lw=0.5, alpha=0.7, shrinkA=5),
            expand_points=(1.2, 1.5),
            expand_text=(1.2, 1.5),
            force_text=0.5,
            force_points=0.5,
            precision=0.01
            )



def sci_set(font_path:str = None, font_family:str='Arial'):
    import matplotlib.pyplot as plt
    import matplotlib as mpl
    from matplotlib import font_manager
    # download the font files and save in this fold
    mpl.rcParams['pdf.fonttype'] = 42
    font_files = font_manager.findSystemFonts(fontpaths=font_path)
    plt.rc('font', family=font_family)
    for file in font_files:
        font_manager.fontManager.addfont(file) # 读取字体库


import numpy as np
import matplotlib.patches as mpathes
import itertools
import matplotlib.pyplot as plt

def LDblock(C:np.ndarray,ax:plt.Axes,vmin:float=0,vmax:float=1,cmap:str='Greys'):
    '''
    ax.set_xlim(0,n), 0.5 ~ n-0.5 n points, n=C.shape[0]=C.shape[1]
    '''
    cbar_rate = 0.05
    mask = np.triu(np.ones_like(C, dtype=bool), k=0) # 不包括对角线的上三角
    C[mask] = np.nan
    n = C.shape[0]
    # 创建旋转/缩放矩阵 (45度旋转伴随缩放)
    t = np.array([[1, 0.5], [-1, 0.5]])
    # 创建坐标矩阵并变换它
    A = np.dot(np.array([(i[1], i[0]) for i in itertools.product(range(n,-1,-1), range(0,n+1,1))]), t)
    # 绘制
    ax.pcolormesh(A[:,1].reshape(n+1,n+1), A[:,0].reshape(n+1,n+1), np.flipud(C),cmap=cmap,vmin=vmin,vmax=vmax)
    rect = mpathes.Rectangle((n/4,-n-cbar_rate*n),n/2,cbar_rate*n,fill=False, edgecolor='black', linewidth=.5)
    ax.add_patch(rect)
    gradient = np.linspace(vmin, vmax, 10).reshape(1, -1)
    clabel = np.linspace(vmin, vmax, 2).round(2)
    cticks = np.linspace(n/4, 3*n/4, 2)
    for i in range(len(clabel)):
        ax.text(cticks[i],-n,clabel[i],ha='center')
    ax.imshow(gradient, cmap=cmap, aspect='auto', extent=[n/4, 3*n/4, -n - cbar_rate * n, -n], origin='lower')
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_ylim(-n-cbar_rate*n,0)
    ax.set_xlim(0,n)
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.spines['bottom'].set_visible(False)


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from scipy.stats import beta


def ppoints(n)->np.ndarray:
    '''
    生成理论均匀分布分位数
    '''
    return np.arange(1,n+1)/(n+1)#(np.arange(1, n+1) - 0.5) / n

class GWASPLOT:
    def __init__(self,df:pd.DataFrame, chr:str, pos:str, pvalue:str,interval_rate:int=.2):
        '''
        必须输入的项目df: 矩阵, chr: 染色体, pos: SNP位点, pvalue: 显著性
        '''
        df = df[[chr, pos, pvalue]].copy()
        df[chr] = df[chr].astype(int)
        df[pos] = df[pos].astype(int)
        df = df.sort_values(by=[chr,pos])
        chrlist = df[chr].unique()
        self.interval = int(interval_rate*df[pos].max())
        df['x'] = df[pos]
        if len(chrlist)>1:
            for ii in chrlist:
                if ii > 1:
                    df.loc[df[chr]==ii,'x'] = df.loc[df[chr]==ii,'x']+df[df[chr]==ii-1]['x'].max()
        df['x'] = (df[chr]-1)*self.interval+df['x']
        df['y'] = df[pvalue]
        df['z'] = df[chr]
        self.chrlist = chrlist
        self.ticks_loc = df.groupby('z')['x'].median()
        self.df = df.set_index([chr,pos])
        pass
    def manhattan(self, threshold:float=None, color_set:list=[], ax:plt.Axes = None, ignore:list=[]):
        df = self.df.copy()
        df['y'] = -np.log10(df['y'])
        if ax == None:
            fig = plt.figure(figsize=[12,6], dpi=600)
            gs = GridSpec(12, 1, figure=fig)
            ax = fig.add_subplot(gs[0:12,0])
        if len(color_set) == 0:
            color_set = ['black','grey']
        colors = dict(zip(self.chrlist,[color_set[i%len(color_set)] for i in range(len(self.chrlist))]))
        ax.scatter(df[~df.index.isin(ignore)]['x'], df[~df.index.isin(ignore)]['y'], alpha=1, s=4, color=df[~df.index.isin(ignore)]['z'].map(colors),rasterized=True)
        if threshold != None and max(df['y'])>=threshold:
            df_annote = df[df['y']>=threshold]
            ax.scatter(df_annote[~df_annote.index.isin(ignore)]['x'], df_annote[~df_annote.index.isin(ignore)]['y'], alpha=1, s=8, color='red',rasterized=True)
            ax.hlines(y=threshold, xmin=0, xmax=max(df['x']),color='grey', linewidth=1, alpha=1, linestyles='--')
        ax.set_xticks(self.ticks_loc, self.chrlist)
        ax.set_xlim([0-self.interval,max(df['x'])+self.interval])
        ax.set_ylim([0,max(df['y'])+0.1*max(df['y'])])
        ax.set_xlabel('Chromosome')
        ax.set_ylabel('-log$_\mathdefault{10}$(p-value)')
        # ax.spines['right'].set_visible(False)
        # ax.spines['top'].set_visible(False)
        return ax
    def qq(self, ax:plt.Axes = None, model:int='qbeta',ci:int=95):
        '''
        可选: bootstrap 抽样次数, ci置信区间(分位数)
        '''
        df = self.df.copy()
        if ax == None:
            plt.rc('font', family='Times New Roman')
            fig = plt.figure(figsize=[12,6], dpi=600)
            gs = GridSpec(12, 1, figure=fig)
            ax = fig.add_subplot(gs[0:12,0])
        p = df['y'].dropna()
        n = len(p)
        # 计算分位数
        o = -np.log10(sorted(p))
        # 生成理论分位数
        e_p = ppoints(n)
        e_theoretical = -np.log10(e_p) # 对于和性状不关联的位点，其pvalue相当于对均匀分布的随机抽样
        if model is not None:
            xi = np.ceil(10**(-e_theoretical)*n)
            lower = -np.log10(beta.ppf(1 - ci/100,xi,n-xi+1))
            upper = -np.log10(beta.ppf(ci/100,xi,n-xi+1))
            # 绘制置信区间
            ax.fill_between(e_theoretical, lower, upper, color='grey', alpha=0.4)
        # 绘制理论线（y=x）和观测点
        ax.plot([0, min(o.max(),e_theoretical.max())], [0, min(o.max(),e_theoretical.max())], lw=1,)
        ax.scatter(e_theoretical, o, s=1, alpha=0.6,rasterized=True)
        ax.set_xlabel('Expected -log$_\mathdefault{10}$(p-value)')
        ax.set_ylabel('Observed -log$_\mathdefault{10}$(p-value)')
        return ax
    


def cal_PVE(df:pd.DataFrame, N:int, beta:str='beta', se_beta:str='se', maf:str='af', n_miss:str='n_miss'):
    '''
    基于gemma或其他程序获得的GWAS结果计算每个位点的pve值\n
    df: gwas结果矩阵, header 需包含 beta(效应值), se of beta(效应值标准误), MAF(ALT基因频率), n_miss(分析每个位点对应的缺失个体数)\n
    N: 被分析的个体总数\n
    Equation: pve = 2*(beta**2)*MAF(1-MAF)/(2*(beta**2)*MAF(1-MAF)+(se**2)*2*(N-n_miss)*MAF(1-MAF))
    '''
    df['pve'] = 2*np.power(df[beta],2)*df[maf]*(1-df[maf])/(2*np.power(df[beta],2)*df[maf]*(1-df[maf])+np.power(df[se_beta],2)*2*(N-df[n_miss])*((1-df[maf])))
    return df